/*
    busmonitor.c - Busmonitor for EIB PEI16 Linux Driver

    Copyright (C) 2006 Ulrich Krenn <e0305817@stud4.tuwien.ac.at>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/



/**** includes ****/
#include <linux/ioctl.h>    /* _IO */
#include "busmonitor.h"


/**** functions ****/
void freeResources(void);
void bailOut(char *message);
void decode_msg(unsigned char *pBuff, int len);
void signalHandler(int sig);
void bus_alarm(int time, int status);
void f_fsync(int t);
void f_send(char* str, int len);


/**** global variables ****/
int             eib_fd=(-1),alarm_status;        /* eib filedescriptor */
char            *device;                        /* deivce name */
unsigned char   statusBAU;


/**** main ****/
int main(int argc, char **argv){
    int                  len=0;
    unsigned long        vers=0;
    unsigned char        recvBuff[MAX_MSG_LEN];

    /**** initialization ****/
    argc = argc;
    argv = argv;
    device = STANDARDDEVICE;
    alarm_status = 0;
    eib_fd = -1;
    statusBAU = 0;

    memset(&recvBuff[0], 0, sizeof(recvBuff));

    /* signaling */
    signal(SIGINT, signalHandler);              /* ctrl-c */
    signal(SIGALRM, signalHandler);             /* alarm() timeout */

    /**** print intro ****/
    print_flush("\n");
    print_flush("=======================================\n");
    print_flush_v("          Busmonitor V %s \n", VERSION);
    print_flush("=======================================\n\n");

    print_flush("Press Strg-C to end programm.\n\n");

    /**** open driver ****/
    print_flush_v("Opening %s\n\n", device);
    if ((eib_fd = open(device, O_RDWR)) == -1){         /* open file descriptor */
        bailOut("Can't open");
    }

    /**** get driver interface version number ****/
    vers = ioctl(eib_fd, IOCTL_VERSION);
    if (fprintf(stdout, "Driver interface version: %X\n\n", (unsigned int)vers) < 0)
        (void)bailOut("Can't write to stdout!");
    bus_flush();
    if (vers != 0x00010010){
        (void)bailOut("Wrong driver interface version!");
    }

    /**** request BAU mode ****/
    print_flush("Requesting BAU mode \n");
    f_send("\x4C\x01\x00\x60", 4);
    f_fsync(ALARM_TIME);

    /* receive messages */
    F_RECEIVE(
        if ((len == 5) && (recvBuff[0] == 0x4B)){       /* PC_GET_VAL.con */
            statusBAU = recvBuff[4];
            break;
        }
    )


    /**** switch to bus monitor mode ****/
    print_flush("Switching to bus monitor mode\n\n");
    f_send("\x46\x01\x00\x60\x90", 5);
    f_fsync(ALARM_TIME);



    /**** receive messages *** (forever ctrl-c) */
    print_flush("Receiving bus monitor messages:\n\n");
    while(1){    
        if ((len = read(eib_fd, (unsigned char*)&recvBuff, MAX_MSG_LEN)) != -1){
            if (len != 0){
                recvBuff[len] = 0;
                decode_msg((unsigned char*)&recvBuff, len);
            }
            else{
                print_flush("Message len: 0 \n");
            }
        }
        else{
            print_flush_v("Receive Status %d\n", len);
        }
    }/* while */
    bailOut("Program Ended");

    return 0;

}/* main */


/* bail Out */
void bailOut(char *message){

    alarm(0);
    if (errno != 0){
        fprintf(stdout, "Error: %s (strerror: %s) \n\n", message, strerror(errno));
    }
    else{
        fprintf(stdout, "Error: %s \n\n", message);
    }
    freeResources();
    exit(-1);

}/* bailOut */


/* free resources */
void freeResources(void){

    if (eib_fd != -1) {
        close(eib_fd);
        eib_fd = -1;
    }
}


/* signalHandler */
void signalHandler(int sig){
    unsigned char    msg[5] = {0x46, 0x01, 0x00, 0x60, statusBAU};

    switch (sig){

        case SIGINT:        /* ctrl-c */
        case SIGTERM:       /* termination signal */

            print_flush("Restoring BAU mode\n\n");
            f_send((char*)&msg[0], 5);
            f_fsync(ALARM_TIME);

            freeResources();
            exit(0);

            break;

        case SIGALRM:

            switch (alarm_status){
                case ALARM_CANT_WRITE_TO_DRIVER:
                    fprintf(stdout,"Can't send message to BAU!\n");
                    break;
                case ALARM_BAU_DOESNT_ANSWER:
                    fprintf(stdout,"BCU doesn't answer!\n");
                    break;
                default:
                    fprintf(stdout,"Unknown Error!\n");
                    break;
            }
            freeResources();
            exit(-1);

            break;

    }/* switch */

}/* signalHandler */


/* decoding of received messages */
void decode_msg(unsigned char *pBuff, int len){
    T_BUSMON_MSG    *pMsg;
    int             i=0;

    pMsg = (T_BUSMON_MSG*)pBuff;

    /* Received 14 bytes @ 42314 ticks: 49 | 0 | 4A A5 | BC | 11 4 | 0 3 | E2 | 0 80 80 | B7 */
    print_flush_v("Received %d bytes ", len);               /* length */

    if (len >= 4){                                         /* byte 3,4: time stamp */
        print_flush_v("@ %d ticks: ", pMsg->timeStamp);
    }

    for (i=0; i<len; i++){                                  /* message */
        bus_print_v("%X ", pBuff[i]);

        if ((i == 0) || (i == 1) || (i == 3) || (i == 4) || (i == 6) || (i == 8) || (i == 9) || (i == 12)){
            bus_print("| ");
        }
    }
    print_flush("\n");

    if (len >= 1){

        switch(pMsg->m_code){                               /* byte 1: message code */

            case 0x49:                                      /* bus monitor message */
                if (len >= 5){                              /* byte 5: control field */

                    /* 1.1.4 ==> 0/0/3, [not] repeated, priority 00 (highest), routing count 6 */

                    /* L_DATA.req FRAME (10x1xx00) */
                    if (((pMsg->control & MASK_BIT7) != 0) && ((pMsg->control & MASK_BIT6) == 0) &&
                        ((pMsg->control & MASK_BIT4) != 0) && ((pMsg->control & MASK_BIT1) == 0) &&
                        ((pMsg->control & MASK_BIT0) == 0)){

    
                        if (len >= 7){                      /* byte 6,7: source adress (always physical 4,4,8) */
                            print_flush_v("%d.", ((pMsg->srcAddrHigh & 0xf0) >> 4));
                            print_flush_v("%d.", pMsg->srcAddrHigh & 0x0f);
                            print_flush_v("%d", pMsg->srcAddrLow);
                        }

                        if (len >= 10){                     /* byte 8,9,10: destination adress */
    
                            if ((pMsg->byte10 & 0x80) == 0){ /* physical address */
                                print_flush_v(" ==> %d.", ((pMsg->destAddrHigh & 0xf0) >> 4));
                                print_flush_v("%d.", pMsg->destAddrHigh & 0x0f);
                                print_flush_v("%d, ", pMsg->destAddrLow);
                            }
                            else{                           /* group address */
                                print_flush_v(" ==> %d/", ((pMsg->destAddrHigh & 0xf8) >> 3));
                                print_flush_v("%d/", pMsg->destAddrHigh & 0x07);
                                print_flush_v("%d, ", pMsg->destAddrLow);
                            }
                        }

                        if ((pMsg->control & 0x20) == 0){   /* repeated, not repeated */
                            print_flush("not repeated, ");
                        }
                        else{
                            print_flush("repeated, ");
                        }


                        switch((pMsg->control & 0x0C) >> 2){    /* priority bits */
    
                            case 0:
                                print_flush("priority 00 (system), ");
                                break;

                            case 1:
                                print_flush("priority 01 (alarm), ");
                                break;

                            case 2:
                                print_flush("priority 02 (high), ");
                                break;

                            case 3:
                                print_flush("priority 03 (low), ");
                                break;

                            default:        /* can't happen */
                                break;
                        }/* switch */

                        if (len >= 10)                          /* routing count */
                            print_flush_v("routing count %d, ", ((pMsg->byte10 & 0x60) >> 5));

                    }    
                    /* ACKNOWLEDGE FRAME (ack, nak, busy) */
                    else if (((pMsg->control & MASK_BIT5) == 0) && ((pMsg->control & MASK_BIT4) == 0) &&
                            ((pMsg->control & MASK_BIT1) == 0) && ((pMsg->control & MASK_BIT0) == 0)){

                            switch(pMsg->control){              /* type */
                                case 0xCC:
                                    print_flush("ACK\n");
                                    break;

                                case 0x0C:
                                    print_flush("NAK\n");
                                    break;

                                case 0xC0:
                                    print_flush("BUSY\n");
                                    break;

                                default:
                                    print_flush("Control filed: Not supported value!\n");
                                    break;
                            }/* switch */
                    }
                    /* OTHER frame format */
                    else{
                        print_flush_v("Unknowen frame format, Control Frame: %d\n", pMsg->control);
                    }

                }/* if len>=5, control field */


                /* Errors: [framing.... | none ] */
                if (len >= 2){            /* byte 2: status bits */
                    unsigned char bitVal=0,first=0;

                    for (i=3; i<8; i++){
                        bitVal = pMsg->status & (1 << i);

                        switch(i){
                            case 3:
                                if (bitVal != 0){
                                    if (first == 0){
                                        print_flush("Errors:");
                                        first = 1;
                                    }
                                    print_flush(" lost flag");
                                }
                                break;

                            /* bit 4: don't care */

                            case 5:
                                if (bitVal != 0){
                                    if (first == 0){
                                        print_flush("Errors: ");
                                        first = 1;
                                    }
                                    print_flush(" parity error");
                                }
                                break;

                            case 6:
                                if (bitVal != 0){
                                    if (first == 0){
                                        print_flush("Errors: ");
                                        first = 1;
                                    }
                                    print_flush(" bit error");
                                }
                                break;

                            case 7:
                                if (bitVal != 0){
                                    if (first == 0){
                                        print_flush("Errors: ");
                                        first = 1;
                                    }
                                    print_flush(" framing error");
                                }
                                break;
    
                            default:
                                break;
                        }/* switch */
                    }/* for */

                    if (first != 0) print_flush("\n");

                 }/* len > 2 */

                break;

            case 0x00:        /* BAU reset msg, 1 byte, msg code 0x00, (not emi standard!) */
                bailOut("BCU reset message!");
                break;

            default:
                bailOut("Unknowen message code!");
                break;
        }/* switch */

    }/* if len > 1 */

    print_flush("\n\n");

}/* decode_msg */


/* timeout + status */
void bus_alarm(int time, int status){
    alarm(time);
    alarm_status = status;
}/* bus_alarm */



/* wait some time till message is sent */
void f_fsync(int t){
    bus_alarm(t, ALARM_CANT_WRITE_TO_DRIVER);
    if (fsync(eib_fd) != 0){
        alarm(0);
        fprintf(stdout,"Can't send message to BAU!\n");
        fflush(stdout);
        freeResources();
        exit(-1);
    }
    alarm(0);
}/* f_fsync */


void f_send(char* str, int len){
    bus_alarm(ALARM_TIME, ALARM_CANT_WRITE_TO_DRIVER);          /* timeout */
    if (write(eib_fd, str, len) == -1){                         /* (PC_GET_VAL.req) */
        bailOut("Can't write to device");
    }
    alarm(0);   /* stop timeout */
}/* f_send */


